#vider compltement la mmoire
rm(list=ls())

#chargement du package
library(rpart)

#****************************************************************************
#***************** Procdures et fonctions **********************************
#****************************************************************************

#affectation bas sur les couts
#modele est l'arbre fourni par RPART
#mat_cout est la matrice de cots
#donnees correspond  l'chantillon test
affectation <- function(modele,mat_cout,donnees){

  #produit entre le vecteur des probas 
  #et une colonne de la matrice de la matrice de cot
  produit <- function(x,p){
    #somme du produit
    return(sum(p*x))
  }

  #prediction pour un individu 
  #bas sur une matrice de cot
  #et les probas. d'affectation
  prediction <- function(p,mc){
    #recuprer dans un vecteur les couts d'affectation
    caff <- apply(mc,2,produit,p)
    #on renvoie l'indice de celui qui minimise les cots
    return(which.min(caff))
  }
    
  #predict fournit les probabilits d'affectation pour chaque individu
  #avec ce "type" de paramtrage
  pred <- predict(modele,type="prob", newdata = donnees)  
  
  #pour chaque individu, utiliser la prdiction base sur les cots
  ychapeau <- apply(pred,1,prediction,mc=mat_cout)
  
  #renvoyer le tout
  return(ychapeau)
}

#bagging sur les arbres
bagging_rpart <- function(y, formule, donnees, mat_cout, repetition){

  #pour avoir le mme rsultat  chaque excution
  set.seed(15)
 
  #structure de liste qui va contenir tous les modles gnrs
  tous <- NULL
  
  #vecteur de rcupration des cots
  cout <- NULL
  
  #effectif
  n <- nrow(donnees)
  
  #vecteur de poids
  poids <- rep(1/n,n)
  
  #paramtres de l'apprentissage rpart
  parametres <- rpart.control(xval=0,cp=0,maxcompete=0,maxsurrogate=0)

  #apprentissage
  for (k in 1:repetition){
  
    #chantillonnage avec remise, tous les individus ont le mme poids
    echantillon <- sample(1:n,n,T,poids)
    #rcuprer le sous-ensemble d'observations corresp.
    subsample <- donnees[echantillon,]
    #apprentissage
    modele <- rpart(formula = formule, data = subsample, method = "class", control = parametres)
    #collecte dans la liste
    tous[[k]] <- modele
    
    #prdiction du modle sur les individus
    ychapeau <- affectation(modele,mat_cout,donnees)
       
    #calcul du cout
    mc <- table(y,ychapeau)
    cout[k] <- sum(mc*mat_cout)
    
  }  
  #renvoyer les cots associs  chaque itration
  return(list(couts=cout,modeles=tous))
}

#affectation par vote  la majorit simple 
# partir d'une liste de classifieurs
#pour un classifieur individuel, on utilise l'affectation base sur les cots
bagging_affectation <- function(liste,mat_cout,donnees){

  #calculer la frquence des affectations pour un individu
  #et envoyer l'index de celui qui est majoritaire
  pred_on_vote <- function(x){
    #calculer la distribution
    frequence <- table(x)
    #renvoyer l'indice du max
    return(as.integer(names(which.max(frequence))))
  }

  #rcupre une matrice d'affectation l'ensemble  des classifieurs
  allpred <- sapply(liste,affectation,mat_cout,donnees)
  
  #puis le prdiction par le vote  la majorit simple
  ychapeau <- apply(allpred,1,pred_on_vote)
  
  #renvoyer la prdiction
  return(ychapeau)
  
}

#****************************************************************************
#***************** PROGRAMME PRINCIPAL **************************************
#****************************************************************************

#dfinir la matrice de cot
cost.matrix <- matrix(c(-3.0,1.0,1.0,1.0,-6.0,1.0,0.0,0.0,0.0),nrow=3,ncol=3)
print(cost.matrix)

#chargement des donnes
setwd("D:/DataMining/Databases_for_mining/concours-cup-etc/dm-cup-2007")
cup <- read.table(file="donnees-dm-cup-2007-full-dataset.txt",header=T,sep="\t",dec=".")

#partition sur la base de la colonne ID
cup.train <- cup[cup$ID=="train",2:length(cup)]
cup.test <- cup[cup$ID=="test",2:length(cup)]           

#*************************************
#Cration et test d'arbres individuels
#*************************************

#paramtres de rpart
#pas de validation croise, et privilgier un arbre profond
parametres <- rpart.control(xval=0,cp=0,maxcompete=0,maxsurrogate=0)

#cration de l'arbre de dcision
std.modele <- rpart(COUPON ~ ., data = cup.train, method="class", control = parametres)

#prdiction sans prise en compte des cots sur la partie test
std.ypred <- predict(std.modele,newdata = cup.test,type="class")
std.m <- table(cup.test$COUPON,std.ypred)
print(std.m) #affichage de la matrice de confusion en test
print(-sum(std.m*cost.matrix)) #calcul du gain (-1.0 * cout)

#prdiction avec affectation tenant compte des couts
cost.std.ypred <- affectation(std.modele,mat_cout=cost.matrix,donnees=cup.test)
cost.std.m <- table(cup.test$COUPON,cost.std.ypred)
print(cost.std.m)
print(-sum(cost.std.m*cost.matrix))

#************************************
#Cration et test de bagging d'arbres
#************************************

#bagging sur rpart - on cre une liste de "repetition" classifieurs
classifieurs <- bagging_rpart(cup.train$COUPON, COUPON ~ ., cup.train, cost.matrix, repetition = 25)

#affichage des couts pour chaque classifieur
print(-classifieurs$couts)

#affectation avec vote  la majorit
bag.ypred <- bagging_affectation(classifieurs$modeles, cost.matrix, cup.test)
bag.m <- table(cup.test$COUPON,bag.ypred)
print(bag.m)
print(-sum(bag.m*cost.matrix))

